/*
 * Copyright (c) 2020 Oracle and/or its affiliates. All rights reserved.
 * Copyright (c) 2020 IBM Corporation. All rights reserved.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0,
 * or the Eclipse Distribution License v. 1.0 which is available at
 * http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * SPDX-License-Identifier: EPL-2.0 OR BSD-3-Clause
 */

// Contributors:
//     IBM - initial drop
//     03/09/2020 - Jody Grassel
//       - 548814 : EL not Auto Detecting Oracle 19C

package org.eclipse.persistence.platform.database.oracle;

public class Oracle19Platform extends Oracle18Platform {
	public Oracle19Platform() {
        super();
    }
}
